from __future__ import annotations

import enum
import time

import dataclasses


@dataclasses.dataclass
class TelemetryPacket:
  """Container for Tello Telemetry.

  All telemetry points are specified in the `Tello SDK User
  Manual
  <https://dl-cdn.ryzerobotics.com/downloads/Tello/Tello%20SDK%202.0%20User%20Guide.pdf>`_

  Args:
    timestamp: When the Telemetry was recorded.
    pitch: the pitch angle of the Tello.
    roll: the roll angle of the Tello.
    yaw: the yaw angle of the Tello.
    vel_x: drone velocity in the x direction.
    vel_y: drone velocity in the y direction.
    vel_z: drone velocity in the z direction.
    temp_l: Corresponds to templ from the SDK
    temp_h: Corresponds to temph from the SDK
    flight_dist: Corresponds to tof from the SDK.
    height: drone height
    battery: drone battery
    barometer: Corresponds to baro from the SDK.
    time: flight time
    accel_x: drone acceleration in the x direction.
    accel_y: drone acceleration in the y direction.
    accel_z: drone acceleration in the z direction.
  """
  timestamp: float
  pitch: int
  roll: int
  yaw: int
  vel_x: int
  vel_y: int
  vel_z: int
  temp_l: int
  temp_h: int
  flight_dist: float
  height: float
  battery: int
  barometer: float
  time: int
  accel_x: float
  accel_y: float
  accel_z: float

  @classmethod
  def from_data_str(cls, data: str):
    """Constructs a TelemetryPacket from a string sent by the drone.

    The format of the telemetry coming from the drone is given in the Tello SDK
    User Guide.
    """
    # There is an ending semicolon
    key_val_pairs = data.split(';')[:-1]

    data_dict = {}
    for pair in key_val_pairs:
      key, val = pair.split(':')
      data_dict[key] = val

    return TelemetryPacket(
        timestamp=time.time(),
        pitch=int(data_dict['pitch']),
        roll=int(data_dict['roll']),
        yaw=int(data_dict['yaw']),
        vel_x=int(data_dict['vgx']),
        vel_y=int(data_dict['vgy']),
        vel_z=int(data_dict['vgz']),
        temp_l=int(data_dict['templ']),
        temp_h=int(data_dict['temph']),
        height=float(data_dict['h']),
        flight_dist=float(data_dict['tof']),
        battery=int(data_dict['bat']),
        barometer=float(data_dict['baro']),
        time=int(data_dict['time']),
        accel_x=float(data_dict['agx']),
        accel_y=float(data_dict['agy']),
        accel_z=float(data_dict['agz'])
    )


class TelloCommand(enum.Enum):
  """Support commands to send the Tello"""
  SDK_ON = 'command'
  START_VIDEO = 'streamon'
  STOP_VIDEO = 'streamoff'
  TAKE_OFF = 'takeoff'
  LAND = 'land'
  HOVER = 'stop'
  RC = 'rc'
  CHECK_BATTERY = 'battery?'


@dataclasses.dataclass
class CommandPacket:
  """Container for TelloCommands

  Args:
    command: The command name to be sent
    payload: A list of arguments to be sent along with the command name
  """
  command: TelloCommand
  payload: list[str]

  def __str__(self) -> str:
    if len(self.payload) == 0:
      return self.command.value
    return f'{self.command.value} ' + ' '.join(self.payload)


class TelloEvent(enum.Enum):
  """Types of Tello Events"""
  START_CONTROL = 'START_CONTROL'
  STOP_CONTROL = 'STOP_CONTROL'
  DROPPED_PACKET = 'DROPPED_PACKET'



def clip_rc(x: float) -> float:
  """Clip RC commands to the [-100, 100] range.

  Tello RC Commands must have an absolute value less than or equal to 100

  Args:
    x: The input to be clipped

  Returns:
    x but clipped to the range [-100, 100].
  """
  return min(max(x, -100), 100)



from .tello import Tello
from .structs import TelloCommand, TelemetryPacket, TelloEvent
from .experiment import Experiment
from .controller import TelloController

__all__ = [
    "Tello", "TelloCommand", "TelemetryPacket", "Experiment", "TelloController",
    "TelloEvent"
]



from typing import Union

import abc
import dataclasses

import csv
import os
import logging
import pathlib

from tello_control.tello import Tello
from tello_control.structs import TelemetryPacket, TelloCommand

TELEM_FIELDS = [field.name for field in dataclasses.fields(TelemetryPacket)]

LOGGER = logging.getLogger("experiment")


class Experiment(abc.ABC):
  """
  Abstract class which provides a framework to analyze a session with the tello.

  Experiments record the telemetry read from the tello and the commands sent to
  the tello to CSVs which can be later analyzed.

  Attributes:
    tello: The Tello which the experiment is being run with.
  """
  def __init__(self, output_path: Union[str, pathlib.Path]):
    self.tello = Tello()
    self._output_path = pathlib.Path(output_path) if isinstance(
        output_path, str
    ) else output_path

  def setUp(self) -> bool:
    """
    Do all setup related to the experiment.

    At a minimum, this method simply connects to the Tello and returns whether
    or not that connection was successful.

    Returns:
      Whether or not connecting to the tello was successful.
    """
    return self.tello.connect()

  def tearDown(self):
    """
    Destroy any objects created during the experiment.

    At a minimum, this method lands the Tello and disconnects from it.
    """
    LOGGER.info("Initiating Landing")
    self.tello.send_command(TelloCommand.LAND, wait_for_success=True)

    LOGGER.info(f"Experiment complete. Tello battery is {self.tello.battery}%")

    self.tello.disconnect()

  def log_results(self):
    """Write the telemetry trace to the experiment output path"""
    if not os.path.exists(self._output_path):
      os.mkdir(self._output_path)

    with open(self._output_path / "trace.csv", "w") as csv_file:
      writer = csv.DictWriter(csv_file, fieldnames=TELEM_FIELDS)

      writer.writeheader()
      for message in self.tello.telem.history:
        writer.writerow(dataclasses.asdict(message))

  def log_command_history(self):
    """Write the command history to the experiment output path"""
    if not os.path.exists(self._output_path):
      os.mkdir(self._output_path)

    with open(self._output_path / "command-history.csv", "w") as csv_file:
      writer = csv.writer(csv_file)
      writer.writerow(["Timestamp", "Command", "Arg0", "Arg1", "Arg2", "Arg3"])
      for timestamp, cmd in self.tello.command_history:
        writer.writerow([timestamp, cmd.command, *cmd.payload])

  def log_events(self):
    """Write the event history to the experiment output path"""
    if not os.path.exists(self._output_path):
      os.mkdir(self._output_path)

    with open(self._output_path / "event-history.csv", "w") as csv_file:
      writer = csv.writer(csv_file)
      writer.writerow(["Timestamp", "Event"])
      for timestamp, event in self.tello.event_history:
        writer.writerow([timestamp, event.value])

  def run(self):
    """Run the experiment

    Running an experiment entails setting up the experiment, running the main
    function, and tearing down any created resources.
    """
    if not self.setUp():
      LOGGER.error("Did not run experiment because could not connect to drone")
      return

    try:
      self.main()
    except KeyboardInterrupt:
      pass
    finally:
      self.tearDown()

      self.log_results()
      self.log_command_history()
      self.log_events()

  @abc.abstractmethod
  def main(self):
    """Main loop of the experiment

    An experiment is created by overriding this method at minimum
    """
    pass



from __future__ import annotations
from typing import Optional, Any

import logging
import queue
import time

from tello_control.structs import CommandPacket, TelloCommand, TelloEvent
from tello_control.udp_interface import (
    TelemetryInterface, CommandInterface, VideoInterface
)

LOGGER = logging.getLogger("tello")


class Tello:
  """The primary interface between the user and the Tello

  Attributes:
    telem: A UdpInterface for reading Drone Telemetry.
    cmd: A UdpInterface for sending commands and receiving their responses.
    video: A UdpInterface for receiving video frames.
    command_history: A list of command packets sent to the drone.
    event_history: A list of events and timestamps
  """
  def __init__(self, ack_timeout: int = 15):
    self.telem = TelemetryInterface()
    self.cmd = CommandInterface()
    self.video = VideoInterface()
    self.command_history: list[tuple[float, CommandPacket]] = []
    self.event_history: list[tuple[float, TelloEvent]] = []

    self._ack_timeout = ack_timeout

  def send_command(
      self,
      command: TelloCommand,
      *args: Any,
      wait_for_success=False
  ) -> tuple[bool, Optional[str]]:
    """Sends a command to the Tello

    Args:
      command: The TelloCommand to send.
      *args: Any arguments the Tello command takes.
      wait_for_success: Whether or not to wait until the Tello response
    """
    str_args = [str(arg) for arg in args]
    packet = CommandPacket(command, payload=str_args)
    self.cmd.send(packet)
    self.command_history.append((time.time(), packet))

    if not wait_for_success:
      return True, None

    # Block until the drone acknowledges
    start = time.time()
    while True:
      try:
        timestamp, msg = self.cmd.messages.get(timeout=self._ack_timeout)
        if timestamp > start:
          return "error" not in msg.lower(), msg
        else:
          # This is an old message, so ignore it
          continue
      except queue.Empty:
        LOGGER.error(f"Drone did not acknowledge command: {command}")
        return False, None

  @property
  def battery(self) -> int:
    """Check the Tello battery
    """
    ret, msg = self.send_command(
        TelloCommand.CHECK_BATTERY, wait_for_success=True
    )
    if ret and msg is not None:
      return int(msg)
    else:
      raise ValueError("Could not check battery")

  def connect(self) -> bool:
    """Connect to the Tello

    Returns:
      Whether or not the Tello connection was successful
    """
    self.telem.connect()
    self.cmd.connect()
    ret, _ = self.send_command(TelloCommand.SDK_ON, wait_for_success=True)
    return ret

  def stream_video(self):
    """Turn on the video stream"""
    self.send_command(TelloCommand.START_VIDEO, wait_for_success=True)
    self.video.connect()

  def end_video_stream(self):
    """Stop the video stream"""
    self.video.disconnect()
    self.send_command(TelloCommand.STOP_VIDEO, wait_for_success=True)

  def disconnect(self):
    """Disconnect from the tello"""
    self.telem.disconnect()
    self.cmd.disconnect()

  def log_event(self, event: TelloEvent, level: int = logging.INFO):
    """Logs an event to the Tello

    Args:
      event: The event to log
      level: The logging level. This comes straight from the `logging` library.
    """
    message = ""
    event_time = time.time()
    if event == TelloEvent.START_CONTROL:
      message = f"Started Control at {event_time}"
    elif event == TelloEvent.STOP_CONTROL:
      message = f"Stopped Control at {event_time}"
    else:
      message = f"Received event {event.value}"

    LOGGER.log(level, message)
    self.event_history.append((event_time, event))



from __future__ import annotations

import abc
from collections.abc import Iterable
import logging
import threading
import time

from tello_control.structs import TelemetryPacket, TelloCommand, TelloEvent
from tello_control import utils, Tello

LOGGER = logging.getLogger("controller")


class TelloController(abc.ABC):
  """
  An abstract base class for a Tello controller.

  It runs a control step at a specified frequency (default 10 Hz) and processes
  telemetry for use in the control step.
  """
  def __init__(self, tello: Tello, control_frequency: float = 10):
    self.tello = tello
    self._telem_lock = threading.Lock()
    self._pending_telem: list[TelemetryPacket] = []
    self._control_frequency = control_frequency

  def start(self):
    """
    Start running the controller.
    """
    self._control_thread = threading.Thread(target=self._run)
    self._telem_thread = threading.Thread(target=self._process_telem)
    self._stop_event = threading.Event()

    self.start_time = time.time()
    self.tello.log_event(TelloEvent.START_CONTROL)

    self._telem_thread.start()
    self._control_thread.start()

  def stop(self):
    """
    Stop running the controller.

    This method closes both the telemetry thread and the control thread that it
    started when the controller began execution.
    """
    self._stop_event.set()

    self._control_thread.join()
    self._telem_thread.join()

    self.end_time = time.time()
    self.tello.log_event(TelloEvent.STOP_CONTROL)

  def _run(self):
    """
    Target of the control thread.

    It runs a while loop which wakes up at a rate equal to the control frequency
    to run the control step and determine what commands to send to the tello.
    Outputs of the control step are clamped between -100 and 100.
    """
    while not self._stop_event.is_set():
      with self._telem_lock:
        try:
          ux, uy, uz, ua = self.step(self._pending_telem)
          ux = utils.clip_rc(ux)
          uy = utils.clip_rc(uy)
          uz = utils.clip_rc(uz)
          ua = utils.clip_rc(ua)

          self.tello.send_command(TelloCommand.RC, ux, uy, uz, ua)
        except Exception as e:
          LOGGER.error(f"Landing tello due to error: {e}")
          self.stop()

        self._pending_telem = []

      time.sleep(1 / self._control_frequency)

  def _process_telem(self):
    """
    Target of the telemtetry thread.

    It runs a while loop which waits until a telemetry message is available and
    prepares it for use by the control step.
    """
    while not self._stop_event.is_set():
      # Blocks until something is available
      telem = self.tello.telem.messages.get()

      # Acquire the lock to add to the pending telemetry
      with self._telem_lock:
        self._pending_telem.append(telem)

  @abc.abstractmethod
  def step(
      self, telemetry: list[TelemetryPacket]
  ) -> tuple[float, float, float, float]:
    """
    The main control step.

    A Controller is created by overriding at least this method. It is run at the
    frequency the Controller was created with.

    Args:
      telemetry: A list of TelemetryPackets which arrived since the last
        execution.

    Returns:
      A tuple representing the x, y, z, and yaw rc command that should be sent
      to the Tello
    """
    pass

  @abc.abstractmethod
  def compute_reference(
      self, now_time: float, telemetry: list[TelemetryPacket]
  ) -> Iterable[float]:
    """
    Compute the reference that the controller is tracking

    A Controller can optionally override this method as an easy way to create
    the reference signal that the drone is supposed to be tracking.

    Args:
      now_time: The current time
      telemetry: A list of Telemetry packets which arrived since the last
        execution.

    Returns:
      An iterable of floats depending on what the reference represents. The
      implementation of the step function is responsible for knowing how to use
      the result of this function.
    """
    raise NotImplementedError()  # noqa



from __future__ import annotations

from typing import Any, Optional

import logging
import threading
import socket
import time
import queue
import errno

import abc

import cv2

from tello_control import structs

LOGGER = logging.getLogger("tello")


class UdpInterface(abc.ABC):
  """An abstract class for a general UDP Interface

  Attributes:
    messages: A Python Queue which stores messages read from the UDP connection
  """
  def __init__(self, name: str, ip: str, port: int):
    self._name = name
    self._ip = ip
    self._port = port
    self._is_streaming = False

    self.messages = queue.Queue()

  def connect(self) -> None:
    """Connect via UDP.

    This method also creates a thread to continuously wait for and read data
    from the UDP socket.
    """
    LOGGER.info(f"Opening {self._name} UDP Connection")

    self._create_connection()
    self._is_streaming = True

    self._thread = threading.Thread(target=self._receive)
    self._stop_event = threading.Event()
    self._thread.start()

  def disconnect(self) -> None:
    """Disconnect from UDP

    This method stops the thread which is consuming data from the socket.
    """
    if not self._is_streaming:
      raise RuntimeError(
          "Cannot end UDP stream because stream was never\
      started"
      )

    LOGGER.info(f"Closing {self._name} UDP Connection")

    self._stop_event.set()
    self._thread.join()

    self._destroy_connection()
    self._is_streaming = False

  def _receive(self) -> None:
    """Main loop of the UDP Interface which polls data from the UDP interface"""
    while not self._stop_event.is_set():
      msg = self._get_data()
      if msg is not None:
        self.messages.put(msg)

    LOGGER.info(f"Closing {self._name} Receive Thread")

  @abc.abstractmethod
  def _get_data(self) -> Optional[Any]:
    """An abstract method which subclasses use to retrieve data from the
    socket"""
    pass

  @abc.abstractmethod
  def _create_connection(self) -> None:
    """Abstract method which subclasses use to create the UDP socket"""
    pass

  @abc.abstractmethod
  def _destroy_connection(self) -> None:
    """Abstract method which subclasses use to destroy the UDP socket"""
    pass


class UdpSocketInterface(UdpInterface):
  """A UdpInterface representing a traditional UDP socket. Also an abstract
  class"""
  def _create_connection(self) -> None:
    self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    self._socket.setblocking(False)
    self._socket.bind(("", self._port))

  def _destroy_connection(self) -> None:
    self._socket.close()

  def _get_data(self) -> Optional[Any]:
    try:
      data, _ = self._socket.recvfrom(1518)
      data = data.decode("utf8").strip()
    except socket.error as e:
      if e.args[0] != errno.EAGAIN:
        LOGGER.error(f"{self._name} encountered socket error {e}")
      return
    except UnicodeDecodeError as e:
      LOGGER.error(f"Could not decode Tello packet: {e}")
      return

    return self._format_packet(data)

  @abc.abstractmethod
  def _format_packet(self, data: str) -> Any:
    """Format a packet from the UDP data

    Args:
      data: Data read from the UDP socket
    """
    pass


class TelemetryInterface(UdpSocketInterface):
  """A UDP socket which processes Tello telemetry

  Attributes:
    history: A list of TelemetryPackets received over the lifetime of the UDP
      connection
  """
  def __init__(self):
    super().__init__("Telemetry", "192.168.10.1", 8890)
    self.history = []

  def _get_data(self) -> Optional[structs.TelemetryPacket]:
    data = super()._get_data()
    if data is not None:
      self.history.append(data)
    return data

  def _format_packet(self, data: str) -> structs.TelemetryPacket:
    return structs.TelemetryPacket.from_data_str(data)


class CommandInterface(UdpSocketInterface):
  """A UDP socket for sending commands to the Tello and receiving responses"""
  def __init__(self):
    super().__init__("Command", "192.168.10.1", 8889)

  def _get_data(self) -> Optional[Any]:
    data = super()._get_data()
    if data is not None:
      LOGGER.debug(f"Received packet '{data[1]}'")
    return data

  def _format_packet(self, data: str) -> tuple[float, str]:
    return time.time(), data

  def send(self, packet: structs.CommandPacket):
    # TODO: Error of socket is not open
    LOGGER.debug(f"Sending {packet}")
    msg = str(packet).encode(encoding="utf-8")
    _ = self._socket.sendto(msg, (self._ip, self._port))


class VideoInterface(UdpInterface):
  """A UdpInterface using OpenCV to open a video socket to the Tello"""
  def __init__(self):
    super().__init__("Video", "0.0.0.0", 11111)

  def _create_connection(self) -> None:
    self._video_capture = cv2.VideoCapture(f"udp://{self._ip}:{self._port}")

  def _destroy_connection(self) -> None:
    self._video_capture.release()

  def _get_data(self) -> Optional[Any]:
    retval, frame = self._video_capture.read()
    if not retval:
      raise RuntimeError("Video Stream disconnected")

    return frame
